@@ -417,6 +417,7 @@ private:
 
   bool m_is_debug_enabled{ false };
 
+  int kfd_ioctl (unsigned long request, void *args) const;
   int kfd_dbg_trap_ioctl (uint32_t action,
                           kfd_ioctl_dbg_trap_args *args) const;
 
@@ -551,25 +552,40 @@ kfd_driver_t::close_kfd ()
     }
 }
 
+int
+kfd_driver_t::kfd_ioctl (unsigned long request, void *args) const
+{
+  dbgapi_assert (is_valid ());
+
+  int ret;
+  do
+    {
+      /* Retry ioctl call if it's interrupted by a signal.  */
+      ret = ::ioctl (*s_kfd_fd, request, args);
+    }
+  while (ret < 0 && errno == EINTR);
+
+  return ret < 0 ? -errno : ret;
+}
+
 int
 kfd_driver_t::kfd_dbg_trap_ioctl (uint32_t action,
                                   kfd_ioctl_dbg_trap_args *args) const
 {
-  dbgapi_assert (is_valid ());
   dbgapi_assert (m_os_pid);
 
   args->pid = *m_os_pid;
   args->op = action;
 
-  int ret = ::ioctl (*s_kfd_fd, AMDKFD_IOC_DBG_TRAP, args);
-  if (ret < 0 && errno == ESRCH)
+  int ret = kfd_ioctl (AMDKFD_IOC_DBG_TRAP, args);
+  if (ret == -ESRCH)
     {
       /* TODO: Should we tear down the process now, so that any operation
          executed after this point returns an error?  */
       return -ESRCH;
     }
 
-  return ret < 0 ? -errno : ret;
+  return ret;
 }
 
 amd_dbgapi_status_t
@@ -583,7 +599,8 @@ kfd_driver_t::check_version () const
 
   /* Check that the KFD major == IOCTL major, and KFD minor >= IOCTL minor.  */
   kfd_ioctl_get_version_args get_version_args{};
-  if (::ioctl (*s_kfd_fd, AMDKFD_IOC_GET_VERSION, &get_version_args))
+
+  if (kfd_ioctl (AMDKFD_IOC_GET_VERSION, &get_version_args))
     fatal_error ("AMDKFD_IOC_GET_VERSION failed");
 
   const version_t kfd_ioctl_version{ get_version_args.major_version,
@@ -660,7 +677,7 @@ kfd_driver_t::agent_snapshot (os_agent_info_t *snapshots,
       agent_info.private_address_aperture_base = entry.scratch_base;
       agent_info.private_address_aperture_limit = entry.scratch_limit;
       agent_info.location_id = entry.location_id;
-      agent_info.simd_count = entry.simd_count;
+      agent_info.simd_count = entry.simd_count * entry.num_xcc;
       agent_info.max_waves_per_simd = entry.max_waves_per_simd;
       agent_info.vendor_id = entry.vendor_id;
       agent_info.device_id = entry.device_id;
@@ -692,13 +709,14 @@ kfd_driver_t::agent_snapshot (os_agent_info_t *snapshots,
         = entry.debug_prop & HSA_DBG_DISPATCH_INFO_ALWAYS_VALID;
       agent_info.watchpoint_exclusive
         = entry.debug_prop & HSA_DBG_WATCHPOINTS_EXCLUSIVE;
+      agent_info.xcc_count = entry.num_xcc;
 
       if (!agent_info.simd_count || !agent_info.max_waves_per_simd
           || !entry.array_count || !entry.simd_arrays_per_engine)
         fatal_error ("Invalid node properties");
 
       agent_info.shader_engine_count
-        = entry.array_count / entry.simd_arrays_per_engine;
+        = (entry.array_count * entry.num_xcc) / entry.simd_arrays_per_engine;
       agent_info.name
         = pci_device_name (agent_info.vendor_id, agent_info.device_id);
     }
